struct entity_t;
// Different types of call-back functions for each entity:
typedef void (*func_t(struct entity_t *self));
typedef void (*touch_t(struct entity_t *self, struct entity_t *other));

// Entity struct:
typedef struct entity_t {
	// All network-necessary data must come first:
	unsigned short	class;
	// To retrieve actual screen position, divide by 256 ( >> 8)
	signed long		x, y;
	signed long		oldx, oldy;
	signed long		dx, dy;		// Scaled up by 256 ( >> 8)
	
	unsigned char	colflags;	// Collision flags (with map)
	
	unsigned long	flags;
	unsigned char	frame;
	unsigned char	width, height;
	
	unsigned char	buttons, lastbuttons;
	
	signed long		next_think;		// Next time that this entity will execute think()

	signed long		invuln_time;
	signed long		attack_time;	// Next time that this entity is allowed to attack
	signed short	health;
	signed short	max_health;
	unsigned char	weapon;

	// All pointer variables and static data must go at the end:
	unsigned char	index;
	struct entity_t	*owner;
	func_t			*think;
	func_t			*render;
	func_t			*death;
	touch_t			*touch;
	func_t			*collide;
} entity_t;

// This structure MUST be the same as above, excluding pointers:
typedef struct netentity_t {
	// All network-necessary data must come first:
	unsigned short	class;
	// To retrieve actual screen position, divide by 256 ( >> 8)
	signed long		x, y;
	signed long		oldx, oldy;
	signed long		dx, dy;		// Scaled up by 256 ( >> 8)
	
	unsigned char	colflags;	// Collision flags (with map)
	
	unsigned long	flags;
	unsigned char	frame;
	unsigned char	width, height;
	
	unsigned char	buttons, lastbuttons;
	
	signed long		next_think;		// Next time that this entity will execute think()

	signed long		invuln_time;
	signed long		attack_time;	// Next time that this entity is allowed to attack
	signed short	health;
	signed short	max_health;
	unsigned char	weapon;
} netentity_t;


// Flags for the flags property of entities:
#define FLAG_FACERIGHT		0x00000001
#define FLAG_ONGROUND		0x00000002
#define FLAG_LIEDOWN		0x00000004
#define FLAG_LOOKUP			0x00000008
#define FLAG_NOGRAVITY		0x00000010
#define FLAG_NOFRICTION		0x00000020
#define FLAG_DEAD			0x00000040
#define FLAG_BULLETPIERCE	0x00001000

// Flags returned thru check_map(entity_t *) function: (colflags)
#define	BLOCKED_GROUND		0x01
#define BLOCKED_LHS			0x02
#define BLOCKED_RHS			0x04
#define BLOCKED_TOP			0x08

#define CLASS_NONE			0
#define CLASS_PLAYER		1
#define CLASS_GRUNT			2
#define CLASS_BULLET		3

#define	MAX_ENTITIES		64

entity_t		**entities = NULL;		// entities on one screen at a time:
entity_t		*player = NULL, *clientplayer = NULL, *serverplayer = NULL;
unsigned int	time = 0;
long	screen_x = 0, screen_y = 0;

unsigned char	*map = NULL;			// Tile-Map pointer.
unsigned char	horiz_scroll = TRUE;
signed char		grayadjust = 0;

void *safe_malloc(size_t size) {
	void *ptr = NULL;
	ptr = malloc(size);
	return ptr;
};

// Resets all the entity's data to original spawning conditions:
void respawn(entity_t *e) {
	// Initialize all the data:
	e->oldx = e->oldy = e->x = e->y = e->dx = e->dy = 0;
	e->class = CLASS_NONE;
	e->frame = 0;
	e->width = 8; e->height = 8;
	e->flags = FLAG_FACERIGHT;
	e->owner = NULL;
	e->render = NULL;
	e->think = NULL;
	e->touch = NULL;
	e->death = NULL;
	e->collide = NULL;
	e->buttons = e->lastbuttons = 0;
	e->invuln_time = 0;
	e->attack_time = 0;
	e->next_think = 0;
	e->weapon = 0;
};

entity_t *spawn() {
	unsigned char i;
	for (i=0; i<MAX_ENTITIES; i++)
		if (entities[i] == NULL) break;
	if (i == MAX_ENTITIES) return NULL;
	
	entity_t	*e = safe_malloc(sizeof(entity_t));
	if (e == NULL) {
		printf_xy(0, 0, "spawn(): Not enough memory!");	
		return NULL;
	}
	entities[i] = e;
	e->index = i;

	// Initialize all the data:
	respawn(e);

	e->health = 1;
	e->max_health = 1;
	
	if (link_play && is_server) {
		packetheader_t send_hdr;
		send_hdr.type = PACKET_ENTITY_SPAWN;
		send_hdr.length = sizeof(int);
		LIO_SendData(&send_hdr, sizeof(packetheader_t));
		LIO_SendData(&i, sizeof(int));
	};
	
	return e;
};

void kill(entity_t *e) {
	if (e == NULL) return;
	if ((e->index > MAX_ENTITIES) || (e < 0)) return;
	entities[e->index] = NULL;
	free(e);
	e = NULL;
};

void damage(entity_t *target, entity_t *inflictor, entity_t *attacker) {
	// If something doesn't exist, then something went wrong:
	if (target == NULL || inflictor == NULL || attacker == NULL) return;
	if ((inflictor->flags & FLAG_DEAD) == FLAG_DEAD) return;
	// Don't call the death function if he's already dead or invulnerable!!
	if (((target->flags & FLAG_DEAD) == FLAG_DEAD) || (target->invuln_time >= time)) return;
	
	// Take away one life:
	target->health--;
	if (target->death) target->death(target);
};


// Set this to think function for entity to remove it: (e.g. Quake-style)
void sub_remove(struct entity_t *self) {
	if (self == NULL) return;
	kill(self);
};

// Test if the two boxes collide:
bool inside_test(int x1, int y1, int x2, int y2, int w1, int h1, int w2, int h2) {
	return (((x1 >= x2) && (x1 <= x2 + w2) && (y1 >= y2) && (y1 <= y2 + h2)) ||
	        ((x1 + w1 >= x2) && (x1 + w1 <= x2 + w2) && (y1 >= y2) && (y1 <= y2 + h2)) ||
	        ((x1 >= x2) && (x1 <= x2 + w2) && (y1 + h1 >= y2) && (y1 + h1 <= y2 + h2)) ||
	        ((x1 + w1 >= x2) && (x1 + w1 <= x2 + w2) && (y1 + h1 >= y2) && (y1 + h1 <= y2 + h2)))
	||	
	       (((x2 >= x1) && (x2 <= x1 + w1) && (y2 >= y1) && (y2 <= y1 + h1)) ||
	        ((x2 + w2 >= x1) && (x2 + w2 <= x1 + w1) && (y2 >= y1) && (y2 <= y1 + h1)) ||
	        ((x2 >= x1) && (x2 <= x1 + w1) && (y2 + h2 >= y1) && (y2 + h2 <= y1 + h1)) ||
	        ((x2 + w2 >= x1) && (x2 + w2 <= x1 + w1) && (y2 + h2 >= y1) && (y2 + h2 <= y1 + h1)));
};

// This function assumes the pointers are valid and are not NULL:
// Perform this check before calling the function.
bool collide(entity_t *e, entity_t *o) {
	bool test = FALSE;
	short x2, y2, x1, y1;
	
	if (e == NULL || o == NULL) return FALSE;
	
	// Check for a collision between entity E and O:
	// If they're close enough, do a sprite-compare:
	x2 = e->x >> 8;
	x1 = o->x >> 8;
	y2 = e->y >> 8;
	y1 = o->y >> 8;
	
	#define w1 (o->width)
	#define h1 (o->height)
	#define w2 (e->width)
	#define h2 (e->height)
	
	//if (((x2 >= x1) && (x2 <= x1 + w1) && (y2 >= y1) && (y2 <= y1 + h1)) ||
	//    ((x1 >= x2) && (x1 <= x2 + w2) && (y1 >= y2) && (y1 <= y2 + h2)))
	if (inside_test(x1, y1, x2, y2, w1, h1, w2, h2))
	{
		return TRUE;
	}
	
	#undef w1
	#undef h1
	#undef w2
	#undef h2
	
	return test;
};

// Determine location to access at (depending on which way we scroll):
unsigned long map_access( int x, int y ) {
	if (horiz_scroll) return (x * TILE_H) + y;
	return (y * TILE_W) + x;
};

unsigned char check_map(entity_t *curr) {
	bool bot_check, rhs_check, lhs_check, top_check;
	unsigned char total = 0;
	int tx, ty, sx, sy, w, h;
	
	unsigned long access;
	
	// {sx, sy} are the pixel coordinates of the current thing.
	sx = (curr->x >> 8);
	sy = (curr->y >> 8);
	
	// {tx, ty} are the tile-coordinates of the current thing on the map.
	tx = (curr->x >> 11);
	ty = (curr->y >> 11);
	
	w = (curr->width >> 3); h = (curr->height >> 3);
	
	bot_check = FALSE;
	lhs_check = FALSE;
	rhs_check = FALSE;
	top_check = FALSE;

	// Bottom check:
	access = map_access( (tx % TILE_W), ((ty+h) % TILE_H) );
	if (curr->flags & FLAG_FACERIGHT) {
		if ( ((map[access] & 0x80) == 0x80) && 
			 inside_test(sx, sy, tx << 3, (ty + h) << 3, curr->width, curr->height, 8, 8) )
		{
			// Entity collides with the tile...
			bot_check = TRUE;
		};
	} else {
		if ( ((map[access] & 0x80) == 0x80) && 
			 inside_test(sx - curr->width, sy, tx << 3, (ty + h) << 3, curr->width, curr->height, 8, 8) )
		{
			// Entity collides with the tile...
			bot_check = TRUE;
		};
	};
	access = map_access( ((tx + 1) % TILE_W), ((ty+h) % TILE_H) );
	if (curr->flags & FLAG_FACERIGHT) {
		if ( ((map[access] & 0x80) == 0x80) && 
			 inside_test(sx, sy, (tx-w+1) << 3, (ty + h) << 3, curr->width, curr->height, 8, 8) )
		{
			// Entity collides with the tile...
			bot_check = TRUE;
		};
	} else {
		if ( ((map[access] & 0x80) == 0x80) && 
			 inside_test(sx - curr->width, sy, (tx-w+1) << 3, (ty + h) << 3, curr->width, curr->height, 8, 8) )
		{
			// Entity collides with the tile...
			bot_check = TRUE;
		};
	};


	// Left check:
	if ((curr->flags & FLAG_FACERIGHT) == FLAG_FACERIGHT) {
		access = map_access( (tx % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx, sy, (tx) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			lhs_check = TRUE;
		};
		ty++; access = map_access( (tx % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx, sy, (tx) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			lhs_check = TRUE;
		};
		ty--;
	} else {
		access = map_access( ((tx - w + 1) % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx - curr->width, sy, (tx - w + 1) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			lhs_check = TRUE;
		};
		ty++; access = map_access( ((tx - w + 1) % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx - curr->width, sy, (tx - w + 1) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			lhs_check = TRUE;
		};
		ty--;
	};
	// Right check:
	if ((curr->flags & FLAG_FACERIGHT) == FLAG_FACERIGHT) {
		access = map_access( ((tx + w) % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx, sy, (tx + w) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			rhs_check = TRUE;
		};
		ty++; access = map_access( ((tx + w) % TILE_W), (ty % TILE_H) );
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx, sy, (tx + w) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			rhs_check = TRUE;
		};
		ty--;
	} else {
		access = map_access( (tx % TILE_W), (ty % TILE_H) ) ;
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx - curr->width, sy, (tx) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			rhs_check = TRUE;
		};
		ty++; access = map_access( (tx % TILE_W), (ty % TILE_H) ) ;
		if ( ((map[access] & 0x80) == 0x80) &&
			 inside_test(sx - curr->width, sy, (tx) << 3, ty << 3, curr->width, curr->height, 8, 8) )
		{
			rhs_check = TRUE;
		};
		ty--;
	};
	
	total = 0;
	total |= (bot_check ? BLOCKED_GROUND : 0x00);
	total |= (lhs_check ? BLOCKED_LHS : 0x00);
	total |= (rhs_check ? BLOCKED_RHS : 0x00);
	total |= (top_check ? BLOCKED_TOP : 0x00);
	
	return total;
};
